#ifndef BAYESCC_H
#define BAYESCC_H

#include <boost/random.hpp>
#include <boost/random/normal_distribution.hpp>
#include <boost/math/distributions/uniform.hpp>


#include "graph.h"
#include <queue>
#include "evaluation.h"


using namespace std;
using namespace boost;


class NodeRank {

public:

    NodeRank(VertexID _vertexID, double _r) :
            vertexID(_vertexID), r(_r) {}

    bool  operator < (const NodeRank  &nodeRank) const {
        return r < nodeRank.r;

    }

    bool  operator > (const NodeRank  &nodeRank) const {
        return r > nodeRank.r;
    }

    VertexID vertexID;
    double r;

};

typedef priority_queue<NodeRank, vector<NodeRank>,less<vector<NodeRank>::value_type> >  NodeRankPriorityList;  // nodes priority list.


class NodeLabelP {

public:

    NodeLabelP(VertexID _vertexID, double _P) :
            vertexID(_vertexID), P(_P) {}

    bool  operator < (const NodeLabelP  &nodeLabelP) const {
        return P < nodeLabelP.P;

    }

    bool  operator > (const NodeLabelP  &nodeLabelP) const {
        return P > nodeLabelP.P;
    }

    VertexID vertexID;
    double P;

};


class EdgeP {

public:

    EdgeP() {}

    EdgeP(EdgeID _edgeID, double _P) :
            edgeID(_edgeID), P(_P) {}

    bool  operator < (const EdgeP  &edgeP) const {
        return P < edgeP.P;

    }

    bool  operator > (const EdgeP  &edgeP) const {
        return P > edgeP.P;
    }

    EdgeID edgeID;
    double P;

};


class BayesCC
{
public:

    //
    BayesCC(Graph * _graph, double _alpha, double _beta, double _teta);
    ~BayesCC();

    //
    set<VertexID> lookupInitialCandidates();
    bool iterate(set<VertexID> &candidates);
    NodeLabelP labelNode(VertexID vertexID);
    void initLabelPriorsP();
    double naiveBayesP(int labelId, VertexID vertexID);
    void lookupNextCandidates(set<VertexID> &candidates, set<VertexID> &newCandidates);
    void initActiveEdges();
    void updateJointProbabilities();
    void eval();
    void initBayesNodes();
    inline double pcheck(double P);
    int topPrior();
    int majorVoting(VertexID vertexID); // not related to naive-bayes, used only for testing.
    void labelRemainingNodesUsingPriors();
    void buildNodeRankPriorityList(set<VertexID> &candidates, NodeRankPriorityList &nodeRankPriorityList);

    //
    priority_queue<NodeLabelP, vector<NodeLabelP>,less<vector<NodeLabelP>::value_type> > notBayes;  // nodes priority list.
    set<VertexID> notBayesSet; // to be removed. we should use the protecter c member in $notFinal. see http://stackoverflow.com/questions/3666387/c-priority-queue-underlying-vector-container-capacity-resize
    set<VertexID> bayesSet;
    priority_queue<EdgeP, vector<EdgeP>,less<vector<EdgeP>::value_type> > inactiveEdges;
    map< pair<int,int> , double > jointP;
    vector<double> labelPriorsP;

    //
    int iterationCount;
    double K;
    double M;
    double nLabels;
    int nChanges;
    int nLabeled;
    bool changed;

    boost::rand48 rng;

    //
    Graph *graph;
    double alpha;
    double beta;
    double teta;
    double edgeBeginProb; // % of active edges to start with, used when beta<1
    double edgeEndProb; // % of edges to leave disabled at end, used when beta<1
    double edgeActivableCount; // no. of edges that can be activated. The others are left inactive.


};

#endif // BAYESCC_H
