#include "graph.h"

#include <boost/config.hpp>
#include <iostream>
#include <fstream>
#include <string>
#include <boost/tokenizer.hpp>
#include <boost/tuple/tuple.hpp>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/visitors.hpp>
#include <boost/graph/breadth_first_search.hpp>
#include <map>
#include <boost/algorithm/string/predicate.hpp>
#include "common.h"
#include <boost/algorithm/string.hpp>
#include <boost/foreach.hpp>
#include <boost/tokenizer.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/graph/connected_components.hpp>
#include <errno.h>
#include <sys/stat.h>
#include "random.h"

using namespace std;
using namespace boost;


void Graph::removeUnlabeledNodes() {

    //  assert("todo" && 0);

}


Graph::Graph(const char *pathname)
{

    assert(pathname != NULL);

    loadFromFile(pathname);

}


Graph::Graph()
{
    name = "Unknown";
}


unsigned long Graph::getFileSize(const char *pathname) {
    struct stat st;

    if(stat(pathname, &st) != 0)
        FATAL("unable to stat file");

    return st.st_size;
}

void Graph::loadFromFile(const char *pathname) {

    cout << "Loading " << pathname << " ...\n";

    name = string(pathname);


    ifstream f(pathname);
    if (!f)
        FATAL("not found: " << pathname);

    f.seekg (0, ios::end);
    unsigned long size = f.tellg();
    f.seekg (0, ios::beg);
    unsigned long prev_pos = 0;
    unsigned long pos = 0;


    // fseek (f, 0, SEEK_END);
    //  unsigned long size = ftell(f);
    //  fseek(f, 0, SEEK_SET);

    string line;

    int state = 0;

    while(getline(f, line)) {
        pos = (f.tellg()*100) / size;
        if(pos - prev_pos > 5) {
            cout << "[" << pos << "%] loading.. |nodes|=" << num_vertices(g) << " |edges|=" << num_edges(g) << "\n";
            prev_pos = pos;
        }



        //    cout << "[" << ftell(f) / size * 100 << "%] Loading...\n";


        boost::algorithm::trim(line);

        if(line.length() == 0)
            continue;

        if(starts_with(line, "*labels*")) {
            state = 1;
            continue;
        }

        if(starts_with(line, "*nodes*")) {
            state = 2;
            continue;
        }

        if(starts_with(line, "*edges*")) {
            state = 3;
            continue;
        }

        if(starts_with(line, "*eof*")) {
            state = 4;
            break;
        }

        switch(state) {
        case 1: // new label
            parseLabel(line);
            break;
        case 2: // new node
            parseNode(line);
            break;
        case 3: // new edge
            parseEdge(line);
            break;

        default:
            assert("WTF!?" && 0);
            break;

        }

    }

    assert(state == 4);


    cout << "Graph loaded.\n";

    nLabels = labels.size();


}


vector<string> tokenize(string  text) {
    tokenizer< escaped_list_separator<char> > tokens(text);
    tokenizer< escaped_list_separator<char> >::iterator token;
    vector<string> v(tokens.begin(), tokens.end());

    return v;
}


void Graph::parseLabel(string text) {

    vector<string> tokens = tokenize(text);
    assert(tokens.size() == 2);

    int labelId = boost::lexical_cast<int>(tokens[0]);
    string labelString = tokens[1];

    //cout << "label:(" << labelId << ",\"" << labelString << "\")\n";

    labels[labelId] = labelString;


}

void Graph::parseEdge(string text) {

    vector<string> tokens = tokenize(text);
    assert(tokens.size() == 3);

    int node1 = boost::lexical_cast<int>(tokens[0]);
    int node2 = boost::lexical_cast<int>(tokens[1]);
    double P = boost::lexical_cast<double>(tokens[2]);

    //cout << "edge:(" << node1 << "," << node2 << "," << P << ")\n";

    insertEdge(node1, node2, P);

}

Edge & Graph::insertEdge(int nodeId1, int nodeId2, double P) {

    EdgeID eID;
    bool ok;

    boost::tie(eID, ok) = boost::add_edge(nodeIds[nodeId1],nodeIds[nodeId2], g);
    assert(ok);

    Edge &edge =g[eID];

    edge.P = P;
    edge.node1 = nodeId1;
    edge.node2 = nodeId2;
    edge.active = true;

    assert(P <= 1+1e-5);
    assert(P >= -1e-5);

    return edge;
}

void Graph::parseNode(string text) {

    vector<string> tokens = tokenize(text);

    if(tokens.size() != 3)
        cout << "parse node: '" << text << "'\n";

    assert(tokens.size() == 3);

    int id = boost::lexical_cast<int>(tokens[0]);
    string name = tokens[1];
    int labelId = boost::lexical_cast<int>(tokens[2]);

    //cout << "node:(" << id << "," << name << "," << labelId << ")\n";

    insertNode(id, name, labelId);

    /*VertexID vID = boost::add_vertex(g); // vID is the index of a new Vertex
    Vertex &node = g[vID];

    node.id = id;
    node.name = name;
    node.gtLabelId = labelId;
    node.active = true;

    nodeIds[id] = vID;*/
}

Edge &Graph::edge(VertexID id1, VertexID id2) {

    EdgeID edgeID;
    bool found;

    tie(edgeID, found) = boost::edge(id1, id2, g);
    assert(found);
    Edge & edge = g[edgeID];

    return edge;

}

void Graph::dumpToNet(const char *pathname) {

    ofstream f(pathname);

    if(!f) {
        FATAL("Error opening output file '" << pathname << "': " << strerror(errno));
    }

    f << "*vertices " << num_vertices(g) << "\n";

    MyGraph::vertex_iterator vertexIt, vertexEnd;
    boost::tie(vertexIt, vertexEnd) = vertices(g);
    for (; vertexIt != vertexEnd; ++vertexIt){
        VertexID vertexID = *vertexIt; // dereference vertexIt, get the ID
        Vertex & node = g[vertexID];
        f << node.id << " \"" << node.name << "\"\n";
    }

    MyGraph::edge_iterator edgeIt, edgeEnd;
    boost::tie(edgeIt, edgeEnd) = edges(g);

    f << "\n*edges\n";

    for (; edgeIt != edgeEnd; ++edgeIt){
        EdgeID edgeID = *edgeIt;
        Edge & edge = g[edgeID];
        Vertex &node1 = g[nodeIds[edge.node1]];
        Vertex &node2 = g[nodeIds[edge.node2]];
        f << node1.id << " " << node2.id << "\n";
    }

    //    custom_bfs_visitor vis;
    //   breadth_first_search(graph, vertex(0, graph), visitor(vis));


}



class custom_bfs_visitor : public boost::default_bfs_visitor
{
public:

    template < typename Vertex, typename Graph >
            void discover_vertex(Vertex u, const Graph & g) const
    {
        std::cout << u << std::endl;
    }
};


void Graph::printComponents() {

    std::vector<int> component(num_vertices(g));
    int num = connected_components(g, &component[0]);

    cout << "Total number of components: " << num << endl;

}


void Graph::printStats() {

    cout << "Graph stats: " << num_vertices(g) << " nodes, " << num_edges(g) <<  " edges, " << labels.size() << " labels.\n";

    int nodesUnknownLabel = 0;
    int nodesKnownLabel = 0;

    MyGraph::vertex_iterator vertexIt, vertexEnd;
    boost::tie(vertexIt, vertexEnd) = vertices(g);
    for (; vertexIt != vertexEnd; ++vertexIt){
        VertexID vertexID = *vertexIt; // dereference vertexIt, get the ID
        Vertex & node = g[vertexID];
        if(UNKNOWN_LABEL(node.gtLabelId))
            nodesUnknownLabel++;
        else
            nodesKnownLabel++;
    }

    cout << "Ratio of known labels (ground truth): " << nodesKnownLabel << "/" << num_vertices(g) << " (" << (double(nodesKnownLabel))/((double)num_vertices(g))*100.0 << "%)\n";

}

void Graph::print() {


    cout << "Nodes:\n";
    

    MyGraph::vertex_iterator vertexIt, vertexEnd;
    boost::tie(vertexIt, vertexEnd) = vertices(g);
    for (; vertexIt != vertexEnd; ++vertexIt){
        VertexID vertexID = *vertexIt; // dereference vertexIt, get the ID
        Vertex & node = g[vertexID];
        cout << "Node id=" << node.id << " name=\"" << node.name << "\" label=\"" << labels[node.gtLabelId] << "\" (" << node.gtLabelId << ")\n";
    }


    MyGraph::edge_iterator edgeIt, edgeEnd;
    boost::tie(edgeIt, edgeEnd) = edges(g);

    for (; edgeIt != edgeEnd; ++edgeIt){
        EdgeID edgeID = *edgeIt;
        Edge & edge = g[edgeID];
        Vertex &node1 = g[nodeIds[edge.node1]];
        Vertex &node2 = g[nodeIds[edge.node2]];
        cout << "Edge " << node1.id << " -> " << node2.id << " P=" << edge.P << "\n";
    }

    //    custom_bfs_visitor vis;
    //   breadth_first_search(graph, vertex(0, graph), visitor(vis));

}




bool Graph::reachesLabeledNode(VertexID vertexID) {

    Vertex & node = g[vertexID];

    if(node.visited)
        return false;

    node.visited = true;

    //assert (node.labelId == 0); // this is probably what we assume.

    MyGraph::adjacency_iterator neighbourIt, neighbourEnd;
    boost::tie(neighbourIt, neighbourEnd) = adjacent_vertices( vertexID, g );

    for (; neighbourIt != neighbourEnd; ++neighbourIt){
        VertexID neighborID = *neighbourIt;        
        Vertex & neighbor = g[neighborID];
        
        if(KNOWN_LABEL(neighbor.labelId))
            return true;

        if(reachesLabeledNode(neighborID))
            return true;
        
    }

    return false;

}


void Graph::cleanVisitedFlag() {

    MyGraph::vertex_iterator vertexIt, vertexEnd;
    boost::tie(vertexIt, vertexEnd) = vertices(g);
    for (; vertexIt != vertexEnd; ++vertexIt){
        VertexID vertexID = *vertexIt; // dereference vertexIt, get the ID
        Vertex & node = g[vertexID];
        node.visited = false;
    }

}


double Graph::meanEdgeP() {

    double mean = 0;

    MyGraph::edge_iterator edgeIt, edgeEnd;
    boost::tie(edgeIt, edgeEnd) = edges(g);

    for (; edgeIt != edgeEnd; ++edgeIt){
        EdgeID edgeID = *edgeIt;
        Edge & edge = g[edgeID];
        mean += edge.P;

    }

    mean /= ((double)num_edges(g));

    return mean;
}


Vertex &Graph::nodeVID(VertexID id) {
    return g[id];
}


Vertex &Graph::nodeFID(int id) {
    return g[nodeIds[id]];
}


void Graph::setLabeledRatio(double ratio) {

    double countLabeled = 0;
    double countUnlabeled = 0;

    vector<VertexID> labeled;

    MyGraph::vertex_iterator vertexIt, vertexEnd;
    boost::tie(vertexIt, vertexEnd) = vertices(g);
    for (; vertexIt != vertexEnd; ++vertexIt){
        VertexID vertexID = *vertexIt; // dereference vertexIt, get the ID
        Vertex & node = g[vertexID];

        if(node.gtLabelId == 0) {
            countUnlabeled++;
        } else {
            countLabeled++;
            labeled.push_back(vertexID);
        }
    }

    random_shuffle(labeled.begin(), labeled.end());

    assert(countLabeled / (((double)num_vertices(g))) >= ratio);

    int countToUnlabel = countLabeled - (((double)num_vertices(g)) * ratio);

    assert(countToUnlabel >= 0);

    cout << "Ratio of known labels will be: " << (int)((countLabeled - countToUnlabel)/((double)num_vertices(g))*100.0) << "%\n";

    for(vector<VertexID>::iterator iter = labeled.begin(); countToUnlabel > 0 && iter != labeled.end(); iter++) {
        VertexID vertexID = *iter;
        Vertex & node = g[vertexID];
        node.gtLabelId = 0;
        countToUnlabel--;
    }


}



Graph *Graph::sample(double ratioNodes) {


    Graph *graph = new Graph();
    graph->copyLabels(this);
    graph->nLabels = nLabels;

    // iterate nodes, add randomly ratioNodes nodes.
    MyGraph::vertex_iterator vertexIt, vertexEnd;
    boost::tie(vertexIt, vertexEnd) = vertices(g);
    for (; vertexIt != vertexEnd; ++vertexIt){

        if(!Random::getTrue(ratioNodes))
            continue;

        VertexID vertexID = *vertexIt; // dereference vertexIt, get the ID
        Vertex & node = g[vertexID];

        Vertex &newNode = graph->insertNode(node.id, node.name, node.gtLabelId);
        newNode.labelId = node.labelId;
        newNode.training = node.training;
        newNode.eLabelId = node.eLabelId;

    }

    // iterate edges, add edges s.t. nodes have been added.
    MyGraph::edge_iterator edgeIt, edgeEnd;
    boost::tie(edgeIt, edgeEnd) = edges(g);

    for (; edgeIt != edgeEnd; ++edgeIt){
        EdgeID edgeID = *edgeIt;
        Edge & edge = g[edgeID];
        if(graph->nodeIdExists(edge.node1) && graph->nodeIdExists(edge.node2)) {
            graph->insertEdge(edge.node1, edge.node2, edge.P);
        }

    }

    return graph;

}



Vertex &Graph::insertNode(int id, string name, int gtLabelId) {

    VertexID vID = add_vertex(g); // vID is the index of a new Vertex
    Vertex &node = g[vID];

    node.aid = num_vertices(g);
    node.id = id;
    node.name = name;
    node.gtLabelId = gtLabelId;
    node.active = true;

    nodeIds[id] = vID;
    
    return node;
}


void Graph::copyLabels(Graph *graph) {

    labels = graph->labels;

}



bool Graph::nodeIdExists(int nodeId) {

    map<int,VertexID>::iterator iter = nodeIds.find(nodeId);
    if(iter == nodeIds.end())
        return false;

    return true;
}


void Graph::setRemoveRatioEdges(double ratio) {

}
