#include "histogram.h"
#include <assert.h>
#include "random.h"


#include <boost/random.hpp>
#include <boost/random/discrete_distribution.hpp>

Histogram::Histogram(int _nBuckets) : nBuckets(_nBuckets)
{
    buckets = (double *) malloc(sizeof(double) * nBuckets);
    memset(buckets, 0, sizeof(double) * nBuckets);

    bucketsNormalized = (double *) malloc(sizeof(double) * nBuckets);
    memset(bucketsNormalized, 0, sizeof(double) * nBuckets);

}


int Histogram::sample() {

    if(normalize() == 0)
        return 0; // all buckets zero, return first bucket.



    rand48 rng = Random::getInstance();

    vector<double> weights;

    for(int i = 0; i < nBuckets; i++)
    weights.push_back(bucketsNormalized[i]);

    boost::random::discrete_distribution<> dist(weights);

    int ret = dist(rng);



    /*cout << "distribution: ";
    for(int i = 0; i < nBuckets; i++)
        cout << i << ":" << bucketsNormalized[i] << ", ";
    cout << "  ret= " << ret << "\n";*/


    return ret;

}


int Histogram::max() {

    int maxBucket = 0;
    double maxValue = 0;

    for(int i = 0; i < nBuckets; i++)  {
        if(buckets[i] > maxValue) {
            maxBucket = i;
            maxValue = buckets[i];
        }
    }

    return maxBucket;

}



double Histogram::normalize() {
    double sum = 0 ;

    for(int i = 0; i < nBuckets; i++) {
        sum+= buckets[i];
    }


    // normalize:
    for(int i = 0; i < nBuckets; i++) {
        bucketsNormalized[i]= buckets[i] / sum;
    }

    return sum;
}


void Histogram::add(Histogram &hist) {

    assert(nBuckets == hist.nBuckets);

    for(int i = 0; i < nBuckets; i++) {
        buckets[i]+= hist.buckets[i];
    }

    // we DO NOT normalize the histogram.

}



Histogram::~Histogram() {

    free(buckets);
    free(bucketsNormalized);

}

