#include "main.h"
#include "common.h"
#include "graph.h"
#include "samplingcc.h"
#include "evaluation.h"
#include "bayescc.h"
#include "bayescc2.h"
#include "bayescc3.h"
#include "bayescc4.h"
#include "bayescc5.h"
#include "bayescc6.h"
#include "bayescc7.h"
#include "wvrn.h"


/* globals */
OPT o;

/* prototypes */
int main(int argc, char **argv);
void initOpt(int argc, char **argv);
void help();
void cleanUp();
void experiment1(Graph &graph);
void experiment2(Graph &graph);
void experiment3(Graph &graph);
void experiment4(Graph &graph);
void experiment5(Graph &graph);
void experiment6(Graph &graph);
void experiment7(Graph &graph);
void experiment8(Graph &graph);
void experiment9(Graph &graph);
void experiment10(Graph &graph);
void experiment11(Graph &graph);
void experiment12(Graph &graph);
void experiment13(Graph &graph);
void experiment14(Graph &graph);
void experiment15(Graph &graph);
void experiment16(Graph &graph);
void experiment18(Graph &graph);
void experiment19(Graph &graph);
void experiment20(Graph &graph);
void experiment21(Graph &graph);
void experiment22(Graph &graph);

/* extern */


/*******************************************/


void experiment1(Graph &graph) {

    double trainingRatio = 2.0/3.0; // use 2/3 (random subset) of labeled nodes for training.
    Evaluation e(&graph);
    e.setupTrainingSet(trainingRatio);

    e.initTrainingSet(); // reset labelId on all nodes. Must be called before running instances of labeling algorithms.

    int nSamples = 100;

    SamplingCC samplingCC(&graph, nSamples);
    samplingCC.eval();

    double samplingCCaccuracy = e.test();

    cout << "samplingCCaccuracy= " << samplingCCaccuracy << "\n";
}


void experiment2(Graph &graph) {

    double trainingRatio = 2.0/3.0;
    Evaluation e(&graph);
    e.setupTrainingSet(trainingRatio);

    // ROUND2DECIMALS avoids problems when varying parameters with loops.
    // for example, if alpha ranges between 0.1 to 1 with steps of 0.05,
    // sometimes there will be decimals after the second positions because
    // of approximations with doubles. With ROUND2DECIMALS we avoid this
    // problem. The algorithms have some optimizations for example when
    // alpha=1, and thay are not activated if alpha=1.00000001 (This value
    // is also invalid since alpha can range only between 0 and 1.)

    e.initTrainingSet();
    double alpha =ROUND2DECIMALS(1.0);
    double beta =ROUND2DECIMALS(1.0);
    double teta = ROUND2DECIMALS(1.0);
    BayesCC bayesCC(&graph, alpha, beta, teta);
    bayesCC.eval();
    double bayesCCaccuracy = e.test();
    cout << "bayesCCaccuracy= " << bayesCCaccuracy << "\n";


}


void experiment3(Graph &graph) {

    double trainingRatio = 2.0/3.0;
    Evaluation e(&graph);
    e.setupTrainingSet(trainingRatio);

    double beta =ROUND2DECIMALS(1.0);
    double teta = ROUND2DECIMALS(1.0);

    for(double alpha = 0.1; alpha <= 1; alpha+=0.1) {

        alpha =ROUND2DECIMALS(alpha);

        e.initTrainingSet();
        BayesCC bayesCC(&graph, alpha, beta, teta);
        bayesCC.eval();
        double accuracy = e.test();
        cout << "#ACC alpha= " << alpha << " beta= " << beta << " teta= " << teta << " accuracy= " << accuracy << "\n";
    }
}


void experiment4(Graph &graph) {

    double trainingRatio = 2.0/3.0;
    Evaluation e(&graph);
    e.setupTrainingSet(trainingRatio);

    double beta =ROUND2DECIMALS(1.0);
    double alpha = ROUND2DECIMALS(1.0);

    for(double teta = 0.1; teta <= 1.05; teta+=0.1) {

        teta =ROUND2DECIMALS(teta);

        e.initTrainingSet();
        BayesCC bayesCC(&graph, alpha, beta, teta);
        bayesCC.eval();
        double accuracy = e.test();
        cout << "#ACC alpha= " << alpha << " beta= " << beta << " teta= " << teta << " accuracy= " << accuracy << "\n";
    }
}


void experiment5(Graph &graph) {

    double trainingRatio = 2.0/3.0;
    Evaluation e(&graph);
    e.setupTrainingSet(trainingRatio);

    double teta =ROUND2DECIMALS(1.0);
    double alpha = ROUND2DECIMALS(1.0);

    for(double beta = 0.1; beta <= 1; beta+=0.1) {

        beta =ROUND2DECIMALS(beta);

        e.initTrainingSet();
        BayesCC bayesCC(&graph, alpha, beta, teta);
        bayesCC.eval();
        double accuracy = e.test();
        cout << "#ACC alpha= " << alpha << " beta= " << beta << " teta= " << teta << " accuracy= " << accuracy << "\n";
    }
}


void experiment6(Graph &graph) {
    assert(false);

}


void experiment7(Graph &graph) {

    double trainingRatio = 2.0/3.0;
    Evaluation e(&graph);
    e.setupTrainingSet(trainingRatio);

    double beta =ROUND2DECIMALS(1.0);
    double alpha = ROUND2DECIMALS(1.0);
    double teta =ROUND2DECIMALS(1.0);

    //for(double teta = 0.1; teta <= 1.05; teta+=0.1) {

    e.initTrainingSet();
    BayesCC2 bayesCC2(&graph, alpha, beta, teta);
    bayesCC2.eval();
    double accuracy = e.test();
    cout << "#ACC alpha= " << alpha << " beta= " << beta << " teta= " << teta << " accuracy= " << accuracy << "\n";
    //  }
}











void experiment8(Graph &graph) {

    double trainingRatio = 2.0/3.0;
    Evaluation e(&graph);
    e.setupTrainingSet(trainingRatio);

    double beta =ROUND2DECIMALS(1.0);
    double teta = ROUND2DECIMALS(1.0);

    for(double alpha = 0.1; alpha <= 1; alpha+=0.1) {

        alpha =ROUND2DECIMALS(alpha);

        e.initTrainingSet();
        BayesCC2 bayesCC2(&graph, alpha, beta, teta);
        bayesCC2.eval();
        double accuracy = e.test();
        cout << "#ACC alpha= " << alpha << " beta= " << beta << " teta= " << teta << " accuracy= " << accuracy << "\n";
    }
}


void experiment9(Graph &graph) {

    double trainingRatio = 2.0/3.0;
    Evaluation e(&graph);
    e.setupTrainingSet(trainingRatio);

    double beta =ROUND2DECIMALS(1.0);
    double alpha = ROUND2DECIMALS(1.0);

    for(double teta = 0.1; teta <= 1; teta+=0.1) {

        teta =ROUND2DECIMALS(teta);

        e.initTrainingSet();
        BayesCC2 bayesCC2(&graph, alpha, beta, teta);
        bayesCC2.eval();
        double accuracy = e.test();
        cout << "#ACC alpha= " << alpha << " beta= " << beta << " teta= " << teta << " accuracy= " << accuracy << "\n";
    }
}


void experiment10(Graph &graph) {

    double trainingRatio = 2.0/3.0;
    Evaluation e(&graph);
    e.setupTrainingSet(trainingRatio);

    double teta =ROUND2DECIMALS(1.0);
    double alpha = ROUND2DECIMALS(1.0);

    for(double beta = 0.1; beta <= 1; beta+=0.1) {

        beta =ROUND2DECIMALS(beta);

        e.initTrainingSet();
        BayesCC2 bayesCC2(&graph, alpha, beta, teta);
        bayesCC2.eval();
        double accuracy = e.test();
        cout << "#ACC alpha= " << alpha << " beta= " << beta << " teta= " << teta << " accuracy= " << accuracy << "\n";
    }
}



void experiment11(Graph &graph) {


    double trainingRatio = 2.0/3.0;
    Evaluation e(&graph);
    e.setupTrainingSet(trainingRatio);

    e.initTrainingSet();

    double alpha =ROUND2DECIMALS(1.0);
    double beta =ROUND2DECIMALS(1.0);
    double teta = ROUND2DECIMALS(1.0);

    BayesCC2 bayesCC2(&graph, alpha, beta, teta);
    e.time.begin();
    bayesCC2.eval();
    double t = e.time.end();

    cout << "#RESULT accuracy= " << e.test() << " time= " << t << "\n";
}



void experiment12(Graph &graph) {

    double trainingRatio = 2.0/3.0;
    Evaluation e(&graph);
    e.setupTrainingSet(trainingRatio);

    e.initTrainingSet();

    double alpha =ROUND2DECIMALS(1.0);
    double beta =ROUND2DECIMALS(1.0);
    double teta = ROUND2DECIMALS(1.0);

    BayesCC2 bayesCC2(&graph, alpha, beta, teta);
    e.time.begin();
    bayesCC2.eval();
    double t = e.time.end();

    cout << "#RESULT accuracy= " << e.test() << " time= " << t << "\n";

}



void experiment13(Graph &graph) {

    double trainingRatio = 2.0/3.0;
    Evaluation e(&graph);
    e.setupTrainingSet(trainingRatio);

    e.initTrainingSet();
    double alpha =ROUND2DECIMALS(1.0);
    double beta =ROUND2DECIMALS(1.0);
    double teta = ROUND2DECIMALS(1.0);
    cout << "#INFO alpha= " << alpha << " beta= " << beta << " teta= " << teta <<"\n";
    BayesCC3 bayesCC3(&graph, alpha, beta, teta);
    bayesCC3.eval();
    double bayesCCaccuracy = e.test();
    cout << "bayesCCaccuracy= " << bayesCCaccuracy << "\n";

}




void experiment14(Graph &graph) {

    double trainingRatio = 2.0/3.0;
    Evaluation e(&graph);
    e.setupTrainingSet(trainingRatio);

    int nSamples = 100;
    int samplesStep = 4;

    e.initTrainingSet();

    for(int i = 1;i <= nSamples;) {

        SamplingCC samplingCC(&graph, i);
        e.time.begin();
        samplingCC.eval();
        unsigned long t = e.time.end();
        double samplingCCaccuracy = e.test();
        cout << "#ACC iterationCount= " << i << " accuracy= " << samplingCCaccuracy << " time= " << t << "\n";

        if(i == 1) {
            i+= samplesStep -1;
        } else{
            i+= samplesStep;
        }

    }


}


void experiment15(Graph &graph) {

    double trainingRatio = 2.0/3.0;
    Evaluation e(&graph);
    e.setupTrainingSet(trainingRatio);

    unsigned long  tMax = 0;

    {
        e.initTrainingSet();
        double alpha =ROUND2DECIMALS(1.0);
        double beta =ROUND2DECIMALS(1.0);
        double teta = ROUND2DECIMALS(1.0);
        BayesCC2 bayesCC2(&graph, alpha, beta, teta);
        e.time.begin();
        bayesCC2.eval();
        unsigned long t = e.time.end();
        cout << "#RESULT2 method= bayescc accuracy= " << e.test() << " time= " << t << "\n";
        tMax = t;
    }

    // tMax = max time that samplingCC may use.

    {
        int nSamplesMax = 2000;
        e.initTrainingSet();
        e.time.begin();
        SamplingCC samplingCC(&graph, nSamplesMax);
        samplingCC.eval(tMax);
        unsigned long t = e.time.end();
        cout << "#RESULT method= samplingcc accuracy= " << e.test() << " time= " << t << "\n";
    }


    {
        e.initTrainingSet();
        BayesCC5 bayesCC5(&graph);
        e.time.begin();
        bayesCC5.eval();
        double bayescc5Acc = e.test();
        double t = e.time.end();
        cout << "#RESULT method= bayescc5 accuracy= " << bayescc5Acc << " time= " << t << "\n";
    }


}

void experiment18(Graph &graph) {

    double trainingRatio = 2.0/3.0;
    Evaluation e(&graph);
    e.setupTrainingSet(trainingRatio);


    e.initTrainingSet();
    BayesCC6 bayesCC6(&graph, o.bayescc6alpha);
    e.time.begin();
    bayesCC6.eval();
    double t = e.time.end();
    double acc = e.test();
    e.printConfusionMatrix();

    e.printAssignedLabels();

    cout << "#RESULT accuracy= " << acc << " time= " << t << "\n";


}



void experiment19(Graph &graph) {

    double trainingRatio = 2.0/3.0;
    Evaluation e(&graph);
    e.setupTrainingSet(trainingRatio);


    e.initTrainingSet();
    WVRN wvRN(&graph);
    wvRN.nIterationsMax = 50;
    e.time.begin();
    wvRN.eval();
    double wvRNAcc = e.test();
    double t = e.time.end();

    cout << "#RESULT accuracy= " << wvRNAcc << " time= " << t << "\n";


}



void experiment20(Graph &graph) {

    double trainingRatio = 2.0/3.0;
    Evaluation e(&graph);
    e.setupTrainingSet(trainingRatio);


    e.initTrainingSet();
    BayesCC6 bayesCC6(&graph, o.bayescc6alpha);
    e.time.begin();
    bayesCC6.eval();
    double bayesCC6Acc = e.test();
    double t = e.time.end();

    cout << "#RESULT accuracy= " << bayesCC6Acc << " time= " << t << "\n";
}



void experiment21(Graph &graph) {

    double trainingRatio = 2.0/3.0;
    Evaluation e(&graph);
    e.setupTrainingSet(trainingRatio);

    e.initTrainingSet();
    BayesCC7 bayesCC7(&graph);
    e.time.begin();
    bayesCC7.eval();
    double acc = e.test();
    double t = e.time.end();

    cout << "#RESULT accuracy= " << acc << " time= " << t << "\n";
}



void experiment22(Graph &graph) {

    double trainingRatio = 2.0/3.0;
    Evaluation e(&graph);
    e.setupTrainingSet(trainingRatio);

    // Bayes
    e.initTrainingSet();
    BayesCC2 bayesCC2(&graph, 1.0, 1.0, 1.0);
    e.time.begin();
    bayesCC2.eval();
    double bayescc2time = e.time.end();
    double bayescc2acc = e.test();

    // Bayes+
    e.initTrainingSet();
    BayesCC6 bayesCC6(&graph, o.bayescc6alpha);
    e.time.begin();
    bayesCC6.eval();
    double bayescc6time = e.time.end();
    double bayescc6acc = e.test();

    // wvRN
    e.initTrainingSet();
    WVRN wvRN(&graph);
    e.time.begin();
    wvRN.eval(bayescc2time);
    double wvRNtime = e.time.end();
    double wvRNacc = e.test();

    // wvRN unbounded
    e.initTrainingSet();
    WVRN wvRNU(&graph);
    e.time.begin();
    wvRNU.eval();
    double wvRNUtime = e.time.end();
    double wvRNUacc = e.test();

    // Sampling
    int nSamplesMax = 200;
    e.initTrainingSet();
    e.time.begin();
    SamplingCC samplingCC(&graph, nSamplesMax);
    samplingCC.eval(bayescc2time);
    double Stime = e.time.end();
    double Sacc = e.test();

    cout << "#RESULT ratioLabeled= " << o.ratioLabeled << " bayesAcc= " << bayescc2acc << " bayespAcc= "
            << bayescc6acc << " wvRNacc= " <<  wvRNacc << " Sacc= " << Sacc <<  " wvRNUacc= " << wvRNUacc << "\n";

    cout << "#RESULT2 ratioLabeled= " << o.ratioLabeled << " bayestime= " << bayescc2time << " bayesptime= "
            << bayescc6time << " wvRNtime= " <<  wvRNtime << " Stime= " << Stime << " wvRNUtime= " << wvRNUtime << "\n";

}



void experiment16(Graph &graph) {

    double trainingRatio = 2.0/3.0;
    Evaluation e(&graph);
    e.setupTrainingSet(trainingRatio);

    double alpha =ROUND2DECIMALS(1.0);
    double beta =ROUND2DECIMALS(1.0);
    double teta = ROUND2DECIMALS(1.0);    

    e.initTrainingSet();
    BayesCC2 bayesCC2(&graph, alpha, beta, teta);
    e.time.begin();
    bayesCC2.eval();
    double bayescc2Time = e.time.end();
    double bayescc2Acc = e.test();

    e.initTrainingSet();
    BayesCC5 bayesCC5(&graph);
    e.time.begin();
    bayesCC5.eval();
    double bayescc5Time = e.time.end();
    double bayescc5Acc = e.test();


    cout << "#RESULT method= bayescc2 accuracy= " << bayescc2Acc << " time= " << bayescc2Time << "\n";
    //   cout << "#RESULT method= bayescc4 accuracy= " << bayescc4Acc << " time= " << bayescc4Time << "\n";
    cout << "#RESULT method= bayescc5 accuracy= " << bayescc5Acc << " time= " << bayescc5Time << "\n";


}



void logDateTime() {
    // Current date/time based on current system
    time_t now = time(0);

    // Convert now to tm struct for local timezone
    tm* localtm = localtime(&now);
    cout << "*** Local date/time: " << asctime(localtm) << " ***\n";

}


int main(int argc, char **argv) {

    cout.precision(4);
    cerr.precision(4);
    std::cout.setf(std::ios::unitbuf);


    // init program options.
    initOpt(argc, argv);

    logDateTime();

    // sanity checks.
    assert(!o.experimentIds.empty());
    assert(o.graphPathname != NULL);

    // load graph.
    Graph graph(o.graphPathname);
    graph.printStats();

    if(o.ratioLabeled != -1) {
        graph.setLabeledRatio(o.ratioLabeled);
    }

    for(set<int>::iterator iter = o.experimentIds.begin(); iter != o.experimentIds.end(); iter++) {

        int experimentId = *iter;

        for(int runId = 0; runId < o.nRuns; runId++) {
            cout << "#CLEAR\n";
            cout << "#INFO runId= " << runId << " experimentId="  << experimentId <<"\n";

            switch(experimentId) {
            case 1:
                experiment1(graph);
                break;

            case 2:
                experiment2(graph);
                break;

            case 3:
                experiment3(graph);
                break;

            case 4:
                experiment4(graph);
                break;

            case 5:
                experiment5(graph);
                break;

            case 6:
                experiment6(graph);
                break;

            case 7:
                experiment7(graph);
                break;

            case 8:
                experiment8(graph);
                break;

            case 9:
                experiment9(graph);
                break;

            case 10:
                experiment10(graph);
                break;

            case 11:
                experiment11(graph);
                break;

            case 12:
                experiment12(graph);
                break;

            case 13:
                experiment13(graph);
                break;


            case 14:
                experiment14(graph);
                break;

            case 15:
                experiment15(graph);
                break;

            case 16:
                experiment16(graph);
                break;

            case 18:
                experiment18(graph);
                break;

            case 19:
                experiment19(graph);
                break;

            case 20:
                experiment20(graph);
                break;

            case 21:
                experiment21(graph);
                break;


            case 22:
                experiment22(graph);
                break;


            default:
                FATAL("unknown experimentId: " << experimentId);
            }

        }
    }

    cleanUp();

}



void  initOpt(int argc, char **argv) {
    int            c;

    cout << "*** cmd: ";
    for(int i = 0; i < argc; i++) {
        cout << argv[i] << " ";
    }
    cout << "***\n";

    o.graphPathname = NULL;
    o.experimentIds.clear();
    o.nRuns = 1;
    o.ratioLabeled = -1;
    o.bayescc6alpha = 1;

    if (argc ==1){
        help();
    }

    opterr = 0;

    while ((c = getopt(argc, argv, "r:e:R:L:a:")) != EOF)
        switch (c)
        {

        case 'r':
        assert(optarg != NULL);
        o.graphPathname = strdup(optarg);
        break;

        case 'R':
        o.nRuns = atoi(optarg);
        break;

        case 'e':        
            o.experimentIds.insert(atoi(optarg));
            break;

        case 'L':
            o.ratioLabeled = strtod(optarg,NULL);
            break;

        case 'a':
            o.bayescc6alpha = strtod(optarg,NULL);
            break;

        default:
            FATAL("option '" << (char) optopt << "' invalid");
        }

}



void help() {

}


void cleanUp() {
    SAFE_FREE(o.graphPathname);
}
